MODULE pmc_client

!--------------------------------------------------------------------------------!
! This file is part of PALM.
!
! PALM is free software: you can redistribute it and/or modify it under the terms
! of the GNU General Public License as published by the Free Software Foundation,
! either version 3 of the License, or (at your option) any later version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2015 Leibniz Universitaet Hannover
!--------------------------------------------------------------------------------!
!
! Current revisions:
! ------------------
!
!
! Former revisions:
! -----------------
! $Id: pmc_client.f90 1765 2016-02-28 13:30:40Z raasch $
!
! 1764 2016-02-28 12:45:19Z raasch
! cpp-statement added (nesting can only be used in parallel mode),
! all kinds given in PALM style
!
! 1762 2016-02-25 12:31:13Z hellstea
! Initial revision by K. Ketelsen
!
! Description:
! ------------
!
! Client part of Palm Model Coupler
!------------------------------------------------------------------------------!

#if defined( __parallel )

    use, intrinsic :: iso_c_binding

#if defined( __lc )
    USE MPI
#else
    INCLUDE "mpif.h"
#endif
    USE  kinds
    USE  PMC_general,   ONLY: ClientDef, DA_NameDef, DA_Namelen, PMC_STATUS_OK, PMC_DA_NAME_ERR, PeDef, ArrayDef, &
                                         DA_Desclen, DA_Namelen, PMC_G_SetName, PMC_G_GetName
    USE  PMC_handle_communicator,   ONLY: m_model_comm,m_model_rank,m_model_npes, m_to_server_comm
    USE  PMC_MPI_wrapper,           ONLY: PMC_Send_to_Server, PMC_Recv_from_Server, PMC_Time,                     &
                                              PMC_Bcast, PMC_Inter_Bcast, PMC_Alloc_mem
    IMPLICIT none
    PRIVATE
    SAVE

!   data local to this MODULE
    Type(ClientDef)                       :: me
!-- TO_DO: what is the meaning of this? Could variables declared in this module
!--        also have single precision?
!    INTEGER, PARAMETER                    :: dp = wp

    INTEGER, save                         :: myIndex = 0                !Counter and unique number for Data Arrays

    ! INTERFACE section

    INTERFACE PMC_ClientInit
        MODULE procedure PMC_ClientInit
    END INTERFACE PMC_ClientInit

    INTERFACE PMC_Set_DataArray_Name
        MODULE procedure PMC_Set_DataArray_Name
        MODULE procedure PMC_Set_DataArray_Name_LastEntry
    END INTERFACE PMC_Set_DataArray_Name

    INTERFACE PMC_C_Get_2D_index_list
        MODULE procedure PMC_C_Get_2D_index_list
    END INTERFACE PMC_C_Get_2D_index_list

    INTERFACE PMC_C_GetNextArray
        MODULE procedure PMC_C_GetNextArray
    END INTERFACE PMC_C_GetNextArray

    INTERFACE PMC_C_Set_DataArray
        MODULE procedure PMC_C_Set_DataArray_2d
        MODULE procedure PMC_C_Set_DataArray_3d
    END INTERFACE PMC_C_Set_DataArray

    INTERFACE PMC_C_setInd_and_AllocMem
        MODULE procedure PMC_C_setInd_and_AllocMem
    END INTERFACE PMC_C_setInd_and_AllocMem

    INTERFACE PMC_C_GetBuffer
        MODULE procedure PMC_C_GetBuffer
    END INTERFACE PMC_C_GetBuffer

    INTERFACE PMC_C_PutBuffer
        MODULE procedure PMC_C_PutBuffer
    END INTERFACE PMC_C_PutBuffer

    ! Public section

    PUBLIC PMC_ClientInit , PMC_Set_DataArray_Name, PMC_C_Get_2D_index_list
    PUBLIC PMC_C_GetNextArray, PMC_C_Set_DataArray
    PUBLIC PMC_C_setInd_and_AllocMem , PMC_C_GetBuffer, PMC_C_PutBuffer ! ,PMC_C_GetServerType

CONTAINS

    SUBROUTINE PMC_ClientInit
        IMPLICIT none

        INTEGER                       :: i
        INTEGER                       :: istat


        ! Tailor MPI environment

        me%model_comm = m_model_comm
        me%inter_comm = m_to_server_comm

        ! Get rank and size
        CALL MPI_Comm_rank (me%model_comm, me%model_rank, istat);
        CALL MPI_Comm_size (me%model_comm, me%model_npes, istat);
        CALL MPI_Comm_remote_size (me%inter_comm, me%inter_npes, istat);

        ! intra communicater is used for MPI_Get
        CALL MPI_Intercomm_merge (me%inter_comm, .true., me%intra_comm, istat);
        CALL MPI_Comm_rank (me%intra_comm, me%intra_rank, istat);

        ALLOCATE (me%PEs(me%inter_npes))

        do i=1,me%inter_npes
           NULLIFY(me%PEs(i)%Arrays)
        end do

        if(me%model_rank == 0) write(0,'(a,5i6)') 'PMC_ClientInit ',me%model_rank,me%model_npes,me%inter_npes,me%intra_rank

        return
    END SUBROUTINE PMC_ClientInit

    SUBROUTINE PMC_Set_DataArray_Name (ServerArrayDesc, ServerArrayName, ClientArrayDesc, ClientArrayName, istat, LastEntry)
        IMPLICIT none
        character(len=*),INTENT(IN)           :: ServerArrayName
        character(len=*),INTENT(IN)           :: ServerArrayDesc
        character(len=*),INTENT(IN)           :: ClientArrayName
        character(len=*),INTENT(IN)           :: ClientArrayDesc
        INTEGER,INTENT(OUT)                   :: istat
        LOGICAL,INTENT(IN),optional           :: LastEntry

        !-- local variables
        type(DA_NameDef)                      :: myName
        INTEGER                               :: myPe
        INTEGER                               :: my_AddiArray=0

        istat = PMC_STATUS_OK
        if(len(trim(ServerArrayName)) > DA_Namelen .or.                                         &
            len(trim(ClientArrayName)) > DA_Namelen)  then !Name too long
            istat = PMC_DA_NAME_ERR
        end if

        if(m_model_rank == 0) then
            myIndex = myIndex+1
            myName%couple_index  = myIndex
            myName%ServerDesc    = trim(ServerArrayDesc)
            myName%NameOnServer  = trim(ServerArrayName)
            myName%ClientDesc    = trim(ClientArrayDesc)
            myName%NameOnClient  = trim(ClientArrayName)
        end if

        !   Broadcat to all Client PEs

        CALL PMC_Bcast ( myName%couple_index,  0,   comm=m_model_comm)
        CALL PMC_Bcast ( myName%ServerDesc, 0,      comm=m_model_comm)
        CALL PMC_Bcast ( myName%NameOnServer,    0, comm=m_model_comm)
        CALL PMC_Bcast ( myName%ClientDesc, 0,      comm=m_model_comm)
        CALL PMC_Bcast ( myName%NameOnClient,    0, comm=m_model_comm)

        !   Broadcat to all Server PEs

        if(m_model_rank == 0) then
            myPE = MPI_ROOT
        else
            myPE = MPI_PROC_NULL
        endif
        CALL PMC_Bcast ( myName%couple_index, myPE, comm=m_to_server_comm)
        CALL PMC_Bcast ( myName%ServerDesc,   myPE, comm=m_to_server_comm)
        CALL PMC_Bcast ( myName%NameOnServer, myPE, comm=m_to_server_comm)
        CALL PMC_Bcast ( myName%ClientDesc,   myPE, comm=m_to_server_comm)
        CALL PMC_Bcast ( myName%NameOnClient, myPE, comm=m_to_server_comm)

        if(present (LastEntry))   then
            CALL PMC_Set_DataArray_Name_LastEntry ( LastEntry = LastEntry)
        end if

        CALL PMC_G_SetName (me, myName%couple_index, myName%NameOnClient)

        return
    END SUBROUTINE PMC_Set_DataArray_Name

    SUBROUTINE PMC_Set_DataArray_Name_LastEntry (LastEntry)
        IMPLICIT none
        LOGICAL,INTENT(IN),optional           :: LastEntry

        !-- local variables
        type(DA_NameDef)                      :: myName
        INTEGER                               :: myPe

        myName%couple_index  = -1

        if(m_model_rank == 0) then
            myPE = MPI_ROOT
        else
            myPE = MPI_PROC_NULL
        endif
        CALL PMC_Bcast ( myName%couple_index,  myPE, comm=m_to_server_comm)

        return
    END SUBROUTINE PMC_Set_DataArray_Name_LastEntry

    SUBROUTINE PMC_C_Get_2D_index_list
       IMPLICIT none

       INTEGER                                 :: i,j,i2,nr,ierr
       INTEGER                                 :: dummy
       INTEGER                                 :: indWin          !: MPI window object
       INTEGER                                 :: indWin2         !: MPI window object
       INTEGER(KIND=MPI_ADDRESS_KIND)          :: win_size        !: Size of MPI window 1 (in bytes)
       INTEGER(KIND=MPI_ADDRESS_KIND)          :: disp            !: Displacement Unit (Integer = 4, floating poit = 8
       INTEGER,DIMENSION(me%inter_npes*2)      :: NrEle           !: Number of Elements of a horizontal slice
       TYPE(PeDef),POINTER                     :: aPE             !: Pointer to PeDef strzcture
       INTEGER(KIND=MPI_ADDRESS_KIND)          :: WinSize         !: Size of MPI window 2 (in bytes)
       INTEGER,DIMENSION(:),POINTER            :: myInd

!      CALL PMC_C_CGet_Rem_index_list

       win_size = c_sizeof(dummy)
       CALL MPI_Win_create (dummy, win_size, iwp, MPI_INFO_NULL, me%intra_comm, indWin, ierr);
       CALL MPI_Win_fence (0, indWin, ierr)                ! Open Window on Server side
       CALL MPI_Win_fence (0, indWin, ierr)                ! Close Window on Server Side and opem on Client side

       do i=1,me%inter_npes
          disp = me%model_rank*2
          CALL MPI_Get (NrEle((i-1)*2+1),2,MPI_INTEGER,i-1,disp,2,MPI_INTEGER,indWin, ierr)
       end do
       CALL MPI_Win_fence (0, indWin, ierr)    ! MPI_get is Non-blocking -> data in NrEle is not available until MPI_fence CALL

       WinSize = 0
       do i=1,me%inter_npes                         !Allocate memory for index array
         aPE => me%PEs(i)
         i2 = (i-1)*2+1
         nr = NrEle(i2+1)
         if(nr > 0)  then
            ALLOCATE(aPE%locInd(nr))
         else
            NULLIFY (aPE%locInd)
         endif
         WinSize = max(nr,WinSize)                  !Maximum window size
       end do

       ALLOCATE(myInd(2*WinSize))
       WinSize = WinSize*c_sizeof(i)

!      local Buffer used in MPI_Get can but must not be inside the MPI Window
!      Here, we use a dummy for MPI Window because the server PEs do not access the RMA window via MPI_get or MPI_Put

       CALL MPI_Win_create (dummy, 1, iwp, MPI_INFO_NULL, me%intra_comm, indWin2, ierr);

       CALL MPI_Win_fence (0, indWin2, ierr)    ! MPI_get is Non-blocking -> data in NrEle is not available until MPI_fence CALL
       CALL MPI_Win_fence (0, indWin2, ierr)    ! MPI_get is Non-blocking -> data in NrEle is not available until MPI_fence CALL

       do i=1,me%inter_npes
          aPE => me%PEs(i)
          nr = NrEle(i*2)
          if(nr > 0 )  then
             disp = NrEle(2*(i-1)+1)
             CALL MPI_Win_lock (MPI_LOCK_SHARED , i-1, 0, indWin2, ierr)
             CALL MPI_Get (myInd,2*nr,MPI_INTEGER,i-1,disp,2*nr,MPI_INTEGER,indWin2, ierr)
             CALL MPI_Win_unlock (i-1, indWin2, ierr)
             do j=1,nr
                aPE%locInd(j)%i = myInd(2*j-1)
                aPE%locInd(j)%j = myInd(2*j)
             end do
             aPE%NrEle = nr
          else
             aPE%NrEle = -1
          end if
       end do

       CALL MPI_Barrier(me%intra_comm, ierr)   ! Dont know why, but this barrier is necessary before we can free the windows

       CALL MPI_Win_free(indWin, ierr);
       CALL MPI_Win_free(indWin2, ierr);
       DEALLOCATE (myInd)

       return
    END SUBROUTINE PMC_C_Get_2D_index_list

    LOGICAL function PMC_C_GetNextArray (myName)
        character(len=*),INTENT(OUT)               :: myName

        !-- local variables
        INTEGER                      :: MyCoupleIndex
        LOGICAL                      :: MyLast                             !Last Array in List
        character(len=DA_Namelen)    :: loName

        loName = 'NoName '
        MyLast = .true.

        CALL PMC_G_GetName (me, MyCoupleIndex, loName, MyLast)

        myName = trim(loName)

        PMC_C_GetNextArray = .NOT. MyLast                        ! Return true if valid array

        return
    END function PMC_C_GetNextArray

    SUBROUTINE PMC_C_Set_DataArray_2d (array)
       IMPLICIT none
!--    TO_DO: is double precision absolutely required here?
       REAL(kind=dp),INTENT(IN),DIMENSION(:,:)    :: array
       !-- local variables
       INTEGER                           :: NrDims
       INTEGER,DIMENSION (4)             :: dims
       INTEGER                           :: dim_order
       TYPE(c_ptr)                       :: array_adr
       INTEGER                           :: i
       TYPE(PeDef),POINTER               :: aPE
       TYPE(ArrayDef),POINTER            :: ar


       dims = 1

       NrDims    = 2
       dims(1)   = size(array,1)
       dims(2)   = size(array,2)
       dim_order = 2

       array_adr = c_loc(array)

       do i=1,me%inter_npes
          aPE => me%PEs(i)
          ar  => aPE%Arrays
          ar%NrDims    = NrDims
          ar%A_dim     = dims
          ar%dim_order = dim_order
          ar%data      = array_adr
       end do

       return
    END SUBROUTINE PMC_C_Set_DataArray_2d

    SUBROUTINE PMC_C_Set_DataArray_3d (array)
       IMPLICIT none
!--    TO_DO: is double precision absolutely required here?
       REAL(kind=dp),INTENT(IN),DIMENSION(:,:,:)  :: array
       !-- local variables
       INTEGER                           :: NrDims
       INTEGER,DIMENSION (4)             :: dims
       INTEGER                           :: dim_order
       TYPE(c_ptr)                       :: array_adr
       INTEGER                           :: i
       TYPE(PeDef),POINTER               :: aPE
       TYPE(ArrayDef),POINTER            :: ar

       dims = 1

       NrDims    = 3
       dims(1)   = size(array,1)
       dims(2)   = size(array,2)
       dims(3)   = size(array,3)
       dim_order =33

       array_adr = c_loc(array)

       do i=1,me%inter_npes
          aPE => me%PEs(i)
          ar  => aPE%Arrays
          ar%NrDims    = NrDims
          ar%A_dim     = dims
          ar%dim_order = dim_order
          ar%data      = array_adr
       end do

       return
    END SUBROUTINE PMC_C_Set_DataArray_3d

   SUBROUTINE PMC_C_setInd_and_AllocMem

      IMPLICIT none

      INTEGER                                 :: i, ierr
      INTEGER                                 :: arlen, myIndex, tag
      INTEGER(idp)                            :: bufsize                   ! Size of MPI data Window
      TYPE(PeDef),POINTER                     :: aPE
      TYPE(ArrayDef),POINTER                  :: ar
      character(len=DA_Namelen)               :: myName
      Type(c_ptr)                             :: base_ptr
      REAL(kind=wp),DIMENSION(:),POINTER      :: base_array
      INTEGER(KIND=MPI_ADDRESS_KIND)          :: WinSize

      myIndex = 0
      bufsize = 8

      ! First stride, Compute size and set index

      do i=1,me%inter_npes
         aPE => me%PEs(i)
         tag = 200

         do while (PMC_C_GetNextArray (myName))
            ar  => aPE%Arrays

            ! Receive Index from client
            tag = tag+1
            CALL MPI_Recv (myIndex, 1, MPI_INTEGER, i-1, tag, me%inter_comm, MPI_STATUS_IGNORE, ierr)

            if(ar%dim_order == 33) then                    ! PALM has k in first dimension
               bufsize = max(bufsize,ar%A_dim(1)*ar%A_dim(2)*ar%A_dim(3))    ! determine max, because client buffer is allocated only once
            else
               bufsize = max(bufsize,ar%A_dim(1)*ar%A_dim(2))
            end if
            ar%BufIndex = myIndex
          end do
      end do

      ! Create RMA (One Sided Communication) window for data buffer

      CALL PMC_Alloc_mem (base_array, bufsize, base_ptr)
      me%TotalBufferSize = bufsize*wp                          !Total buffer size in Byte

      WinSize = me%TotalBufferSize
!      write(9,'(a,8i7)') 'PMC_S_SetInd_and_Mem ',m_model_rank,me%inter_npes,WinSize,ar%A_dim
      CALL MPI_Win_create (base_array, WinSize, wp, MPI_INFO_NULL, me%intra_comm, me%BufWin, ierr);
      CALL MPI_Win_fence (0, me%BufWin, ierr);                    !  Open Window to set data
      CALL MPI_Barrier(me%intra_comm, ierr)

      do i=1,me%inter_npes
         aPE => me%PEs(i)

         do while (PMC_C_GetNextArray (myName))
            ar  => aPE%Arrays
            ar%SendBuf = base_ptr
         end do
      end do

      return
   END SUBROUTINE PMC_C_setInd_and_AllocMem

   SUBROUTINE PMC_C_GetBuffer (WaitTime)
      IMPLICIT none
      REAL(kind=dp),INTENT(OUT),optional         :: WaitTime

      !-- local variables
      INTEGER                                 :: ip, ij, ierr
      INTEGER                                 :: nr                 ! Number of Elements to getb from server
      INTEGER                                 :: myIndex
      REAL(kind=dp)                           :: t1,t2
      TYPE(PeDef),POINTER                     :: aPE
      TYPE(ArrayDef),POINTER                  :: ar
      INTEGER,DIMENSION(1)                    :: buf_shape
      REAL(kind=wp),POINTER,DIMENSION(:)      :: buf
      REAL(kind=wp),POINTER,DIMENSION(:,:)    :: data_2d
      REAL(kind=wp),POINTER,DIMENSION(:,:,:)  :: data_3d
      character(len=DA_Namelen)               :: myName
      INTEGER(kind=MPI_ADDRESS_KIND)          :: target_disp

      t1 = PMC_Time()
      CALL MPI_Barrier(me%intra_comm, ierr)                         ! Wait for server to fill buffer
      t2 = PMC_Time()
      if(present(WaitTime)) WaitTime = t2-t1

      CALL MPI_Barrier(me%intra_comm, ierr)                         ! Wait for buffer is filled

      do ip=1,me%inter_npes
         aPE => me%PEs(ip)

         do while (PMC_C_GetNextArray (myName))
            ar  => aPE%Arrays
            if(ar%dim_order == 2) then
               nr = aPE%NrEle
            else if(ar%dim_order == 33) then
               nr = aPE%NrEle*ar%A_dim(1)
            end if

            buf_shape(1) = nr
            CALL c_f_pointer(ar%SendBuf, buf, buf_shape)

            if(nr > 0)   then
               target_disp = (ar%BufIndex-1)
               CALL MPI_Win_lock (MPI_LOCK_SHARED , ip-1, 0, me%BufWin, ierr)
               CALL MPI_Get (buf, nr, MPI_REAL, ip-1, target_disp, nr, MPI_REAL, me%BufWin, ierr)
               CALL MPI_Win_unlock (ip-1, me%BufWin, ierr)
            end if

            myIndex = 1
            if(ar%dim_order == 2) then

               CALL c_f_pointer(ar%data, data_2d, ar%A_dim(1:2))
               do ij=1,aPE%NrEle
                  data_2d(aPE%locInd(ij)%j,aPE%locInd(ij)%i) = buf(myIndex)
                  myIndex = myIndex+1
               end do
            else if(ar%dim_order == 33) then
               CALL c_f_pointer(ar%data, data_3d, ar%A_dim(1:3))
               do ij=1,aPE%NrEle
                  data_3d(:,aPE%locInd(ij)%j,aPE%locInd(ij)%i) = buf(myIndex:myIndex+ar%A_dim(1)-1)
                  myIndex = myIndex+ar%A_dim(1)
               end do
            end if

         end do
      end do
      return
   END SUBROUTINE PMC_C_GetBuffer

   SUBROUTINE PMC_C_PutBuffer (WaitTime)
      IMPLICIT none
      REAL(kind=dp),INTENT(OUT),optional         :: WaitTime

      !-- local variables
      INTEGER                                 :: ip, ij, ierr
      INTEGER                                 :: nr                 ! Number of Elements to getb from server
      INTEGER                                 :: myIndex
      REAL(kind=dp)                           :: t1,t2
      TYPE(PeDef),POINTER                     :: aPE
      TYPE(ArrayDef),POINTER                  :: ar
      INTEGER,DIMENSION(1)                    :: buf_shape
      REAL(kind=wp),POINTER,DIMENSION(:)      :: buf
      REAL(kind=wp),POINTER,DIMENSION(:,:)    :: data_2d
      REAL(kind=wp),POINTER,DIMENSION(:,:,:)  :: data_3d
      character(len=DA_Namelen)               :: myName
      INTEGER(kind=MPI_ADDRESS_KIND)          :: target_disp


      do ip=1,me%inter_npes
         aPE => me%PEs(ip)

         do while (PMC_C_GetNextArray (myName))
            ar  => aPE%Arrays
            if(ar%dim_order == 2) then
               nr = aPE%NrEle
            else if(ar%dim_order == 33) then
               nr = aPE%NrEle*ar%A_dim(1)
            end if

            buf_shape(1) = nr
            CALL c_f_pointer(ar%SendBuf, buf, buf_shape)

            myIndex = 1
            if(ar%dim_order == 2) then
               CALL c_f_pointer(ar%data, data_2d, ar%A_dim(1:2))
               do ij=1,aPE%NrEle
                  buf(myIndex) = data_2d(aPE%locInd(ij)%j,aPE%locInd(ij)%i)
                  myIndex = myIndex+1
               end do
            else if(ar%dim_order == 33) then
               CALL c_f_pointer(ar%data, data_3d, ar%A_dim(1:3))
               do ij=1,aPE%NrEle
                  buf(myIndex:myIndex+ar%A_dim(1)-1) = data_3d(:,aPE%locInd(ij)%j,aPE%locInd(ij)%i)
                  myIndex = myIndex+ar%A_dim(1)
               end do
            end if

            if(nr > 0)   then
               target_disp = (ar%BufIndex-1)
               CALL MPI_Win_lock (MPI_LOCK_SHARED , ip-1, 0, me%BufWin, ierr)
               CALL MPI_Put (buf, nr, MPI_REAL, ip-1, target_disp, nr, MPI_REAL, me%BufWin, ierr)
               CALL MPI_Win_unlock (ip-1, me%BufWin, ierr)
            end if

          end do
      end do


      t1 = PMC_Time()
      CALL MPI_Barrier(me%intra_comm, ierr)                         ! Wait for server to fill buffer
      t2 = PMC_Time()
      if(present(WaitTime)) WaitTime = t2-t1

      CALL MPI_Barrier(me%intra_comm, ierr)                         ! Wait for buffer is filled


      return
    END SUBROUTINE PMC_C_PutBuffer

#endif
END MODULE pmc_client
